package ap;

import ap.proof.{ConstraintSimplifier, ProofTree}
import ap.parameters.{Settings, Param}
import ap.util.Debug

object CmdlMain {

  def printGreeting = {
    println("________       _____")                                 
    println("___  __ \\_________(_)________________________________")
    println("__  /_/ /_  ___/_  /__  __ \\  ___/  _ \\_  ___/_  ___/")
    println("_  ____/_  /   _  / _  / / / /__ /  __/(__  )_(__  )")
    println("/_/     /_/    /_/  /_/ /_/\\___/ \\___//____/ /____/")  
  }
  
  def printUsage = {
    println("Usage: princess <option>* <inputfile>*")
    println
    printOptions
  }
  
  def printOptions = {
    println("Options:")
    println("  [+-]printTree                 Output the constructed proof tree        (default: -)")
    println("  -timeout=val                  Set a timeout in milliseconds            (default: infty)")
    println("  [+-]simplifyConstraints       Permanently simplify constraints         (default: +)")
    println("  [+-]assert                    Enable runtime assertions                (default: +)")
    println("  [+-]traceConstraintSimplifier Show the constraint simplifications done (default: -)")
  }
  
  def proveProblems(settings : Settings,
                    problems : Seq[(String, java.io.Reader)],
                    userDefStoppingCond : => Boolean) : unit = {
    if (problems.isEmpty) {
      println("No inputs given, exiting")
      return
    }
    
    Debug.enableAllAssertions(Param.ASSERTIONS(settings))

    try {
          val simplifier = if (Param.SIMPLIFY_CONSTRAINTS(settings)) {
                             if (Param.TRACE_CONSTRAINT_SIMPLIFIER(settings))
                               ConstraintSimplifier.ALL_SIMPLIFIER_TRACE
                             else
                               ConstraintSimplifier.ALL_SIMPLIFIER
                           } else {
                             ConstraintSimplifier.NO_SIMPLIFIER
                           }
          for ((filename, reader) <- problems) {
            val timeBefore = System.currentTimeMillis
            
            println("Loading " + filename + " ...")
            val prover = new IntelliFileProver(reader,
                                               simplifier,
                                               Param.TIMEOUT(settings),
                                               true,
                                               userDefStoppingCond)

            println
            
            prover.result match {
              case IntelliFileProver.Proof(tree, model) => {
                println("Formula is valid, resulting constraint:")
                println("" + tree.closingConstraint)
                if (!model.isTrue) {
                  println
                  println("Concrete witness:")
                  println("" + model)
                }
              }
              case IntelliFileProver.NoProof =>  {
                println("No proof found")
              }
              case IntelliFileProver.CounterModel(model) =>  {
                println("Formula is invalid, found a countermodel:")
                println("" + model)
              }
              case IntelliFileProver.NoCounterModel =>  {
                println("No countermodel exists, formula is valid")
              }
              case IntelliFileProver.SearchCancelled =>  {
                println("Cancelled or timeout")
              }
            }
            
            if (Param.PRINT_TREE(settings)) {
              println
              println("Proof tree:")
              println(prover.proofTree)
            }
            
            val timeAfter = System.currentTimeMillis
            
            println
            println("" + (timeAfter - timeBefore) + "ms")
          }
    } catch {
      case e : Throwable => {
        println("ERROR: " + e.getMessage)
        e.printStackTrace
        return
      }
    }
  }

  def main(args: Array[String]) : unit = {
    printGreeting
    println
    
    val (settings, inputs) = try { Settings.fromArguments(args) } catch {
      case e : Throwable => {
        println(e.getMessage)
        println
        printUsage
        return
      }
    }

    proveProblems(settings,
                  for (name <- inputs.projection)
                  yield (name, new java.io.FileReader(new java.io.File (name))),
                  false)
  }
}
